// le20i.h - LE20I QBUS hosted MISS interface by Vitya Yakimchuk
//
// Mar-2004 by Solo, adapted from Serge sources
//

#ifndef _LE20I_H
#define _LE20I_H

#include "miss.h"

class LE20I: public MISS
{

 public:
  LE20I() {}	
  LE20I(QBUS* host, int c = 0): MISS(host, c)
  {
    //    au();
    //    clear();
  }

  virtual ~LE20I() {}

  // MISS operations
  virtual void sync()  { qbus->read (sync_port ()); }     // Send pulse to NIM output
  virtual void addr()  { set_mode (0x3f & ur); }          // Set address mode
  virtual void au()    { set_mode (0xc0 | (0x3f & ur)); } // Set autonomous mode
  virtual void clear() { set_command (0x8000); set_command (0x8000); } // Send Clear to crate
  virtual short status() { qbus->write (addr_port (), 1); return qbus->read (win_port ()); } // Get status

  virtual int tr ()  {  return ((status () & 0x10) > 0);  }
  virtual int gt ()  {  return ((status () & 0x20) > 0);  }
  virtual int bus_error ()  {  return ((status () & 0x80) > 0);  }

  virtual void print() {
    static const char *status_name[11]={"T1","T2","T3",
					"AR - autonomous mode",
					"TR - active request",
					"GT - ready",
					"ERR - error",
					"BERR - bus error",
					"PCHN - serial numbers read",
					"PCHI - serial data read",
					"IN1 - input request"};
    unsigned short s=(status()&0xffff);
    printf("MISS crate %d status=0x%04x\n",crate,s); 
    for(int i=0; i<11; i++) {
      if((s&0x0001) == 1) printf("0x%04x = %s\n",1<<i,status_name[i]);
      s=s>>1;
    }
  }

  virtual void write(int n, int a, short data) {
    addr ();
    set_data (data);
    set_command (command (n, a, 3));              // Write function 3
    if (bus_error ()) error = bad_address;
    else              error = none;
  }

  virtual short read(int n, int a) {
    short data;
    addr ();
    set_command (command (n, a, 0));              // Read function 0!
    data = get_data ();
    if (bus_error ()) error = bad_address;
    else              error = none;
    return data;
  }

  virtual short function(int n, int a, int f, short data = 0) {
    addr ();
    if ((f == 3) or (f == 5)) set_data (data);
    set_command (command (n, a, f));
    if ((f == 0) or (f == 1) or (f == 4)) data = get_data ();
    if (bus_error ()) error = bad_address;
    else              error = none;
    return data;
  }

  virtual int sdata(int maxsize, short* data, short* num = NULL) {
    int ending;
    int tt=0;

    // wait for GT to disappear
    while (gt ())      
      if (tt++ > timeout)
	{
	  error = gt_timeout;
	  return -1;
	}

    set_mode (0x80 + ur);       //  Set PCHI mode

    int cnt = 0;
    if (tr ())
      {
	set_command (0);         //  Set F(0) function
	do
	  {
	    if (cnt++ >= maxsize)
	      {
		error = buffer_overflow;
		return maxsize;
	      }
	    *(data++) = get_data ();
	    ending = not tr ();
	    if (num != NULL) *(num++) = get_num ();
	    else                        get_num ();
	  } while (not ending);
      }

    set_mode(ur);                // set normal mode

    error = none;
    return cnt;
  }

  virtual int snumber(int maxsize, short* num) {
    int ending;
    
    set_mode (0x40 + ur);       //  Set PCHN mode

    int cnt = 0;
    if (tr ())
      {
	set_command (0);         //  Set F(0) function
	do
	  {
	    if (cnt++ >= maxsize)
	      {
		error = buffer_overflow;
		return maxsize;
	      }
	    ending = not tr ();
	    *(num++) = get_num ();
	  } while (not ending);
      }

    set_mode (ur);              // reset mode back

    return cnt;
  }

  // fast serial readout by blocks of 16 words
  int sdata_fast16(int maxsize, short* data) {
    int ending;
    int tt=0;

    // wait for GT to disappear
    while (gt ())      
      if (tt++ > timeout)
	{
	  error = gt_timeout;
	  return -1;
	}

    set_mode (0x80 + ur);       //  Set PCHI mode

    int cnt = 0;
    if (tr ())
      {
	set_command (0);         //  Set F(0) function
	do
	  {
	    for(int i=0; i<16; i++) {
	      if (cnt++ >= maxsize)
		{
		  error = buffer_overflow;
		  return maxsize;
		}
	      *(data++) = get_data ();
	      get_num();         // has to be done, otherwise crap?
	    }
	    ending = not tr ();
	  } while (not ending);
      }

    set_mode(ur);                // set normal mode

    error = none;
    return cnt;
  }

  void set_mask (short mask)
  {
    qbus->write (addr_port (), 2);
    qbus->write (win_port (), mask);
  }

  int get_mask ()
  {
    qbus->write (addr_port (), 2);
    return qbus->read (win_port ());
  }

 private:

// Private macros
  short addr_port ()  {  return base_port;  }
  short win_port ()   {  return base_port + 2;  }
  short scom_port ()  {  return base_port + 4;  }
  short sync_port ()  {  return base_port + 0xa;  }
  short wdata_port () {  return base_port + 0xa;  }
  short rdata_port () {  return base_port + 0xc;  }

  void set_data (short data)  {  qbus->write (wdata_port (), data);  }
  short get_data ()           {  return qbus->read (rdata_port ());  }
  short get_num ()            {  return qbus->read (scom_port ());  }

  short command (int n, int a, int f)
  {
    return ((f & 7) << 12) + ((n & 0x3f) << 6) + (a & 0x3f);
  }

  void set_command (short command)
  {
    qbus->write (addr_port (), 0);
    qbus->write (win_port (), command);
  }

  void set_mode (short mode)
  {
    qbus->write (addr_port (), 3);
    qbus->write (win_port (), mode);
  }

};

#endif
