// tl2.h - TL2-50 SUMMA trigger module by N.Shalanda 
//
// Oct-2008 by Solo
//
#ifndef _TL2_H
#define _TL2_H

#include "le20c.h"

#define TL2_TRIGMASK  12
#define TL2_SCALER    13
#define TL2_INTENSITY 15

#define TL2_COUNTER(a) (unsigned short)h->function(n,a,2)+((h->function(n,a,4)&0x1fff)<<16)
#define TL2_TRIGGER(a) (unsigned short)h->function(n,a,6)+((h->function(n,a,7)&0x1fff)<<16)

class TL2
{

 public:

  TL2 (LE20C* host, int number)
  {
    n = number;
    h = host;
  }

  ~TL2 () {};

  int number() { return n; }
  LE20C * host() { return h; }

  // reset all
  void reset() {
    h->act(n,0,26); // reset module
    h->act(n,0,28); // reset intensimeter
    h->act(n,1,28); // reset all counters
  }

  // enable/diable automatic busy
  void setbusy(int enable) {
    if(enable!=0) h->act(n,1,26); // enable auto busy
    else h->act(n,1,24);          // disable auto busy
  }

  unsigned int readmask (int x) {
    return ((h->read(n,x)&0xffff)+((h->read(n,4+x)&0xffff)<<16));
  }

  void writemask (int x, unsigned int mask) {
    h->write(n,x,mask&0xffff);         // anti
    h->write(n,4+x,(mask>>16)&0xffff); // coinc
  }

  // T1-T4 mask bits, 1=masked,0=enabled
  int readtrigmask () {
    return h->read(n,TL2_TRIGMASK)&0x0f;
  }

  void writetrigmask (int mask) {
    mask=(mask&0x0f)+(readfreq()<<4); // remember intensimeter freq bits
    h->write(n,TL2_TRIGMASK,mask);
  }

  int readbusy() {
    return (h->read(n,TL2_TRIGMASK)>>4)&0x3;
  }

  // intensity meter frequency divider bits
  int readfreq() {
    return (h->read(n,TL2_TRIGMASK)>>6)&0x3;
  }

  // for writing we have to omit 2 busy bits
  void writefreq(int mask) {
    mask=(mask<<4)+readtrigmask(); // remember trigger mask bits
    h->write(n,TL2_TRIGMASK,mask);
  }

  // prescalers for channels 0-1
  int readpsc(int x) {
    return h->read(n,TL2_SCALER+x)&0xff;
  }

  void writepsc(int x, int s) {
    h->write(n,TL2_SCALER+x,s&0xff);
  }

  // read intensimeter memory
  int intensity() {
    return h->read(n,TL2_INTENSITY)&0xffff;
  }

  // read input counter [0-15], 28 bit, 29bit = overflow
  int counter(int chan) {
    if(chan<0 || chan>15) {
      printf("tl2::counter chan %d is out of [0,15] range!\n",chan);
      return -1;
    }
    else return TL2_COUNTER(chan);
  }

  // read trigger counters for all
  int trigall(int chan) { 
    if(chan<0 || chan>4) {
      printf("tl2::trigall chan %d is out of [0,4] range!\n",chan);
      return -1;
    }   
    return TL2_TRIGGER(chan);
  }

  // read trigger counters blocked by busy
  int trigbusy(int chan) {
    if(chan<0 || chan>4) {
      printf("tl2::trigbusy chan %d is out of [0,4] range!\n",chan);
      return -1;
    }
    return TL2_TRIGGER(chan+4);
  }

  // read summary trigger
  int trigsum() {
    return TL2_TRIGGER(8);
  }

  // read busy counter
  int busysum() {
    return TL2_TRIGGER(9);
  }

  // print 2bit formatted mask
  void printmaskbit(unsigned int mask, int bit) {
    char TL2_MASKNAME[4]={'O','C','A','X'}; // mask name
    int i=((mask>>(16+bit))&1)+2*((mask>>bit)&1);
    printf("%c",TL2_MASKNAME[i]);
  }

  // print formatted contents
  void print  () {
    printf("TL2 module %d contents:\n",n);
    for(int x=0; x<4; x++) {
      unsigned int m=readmask(x);
      printf("mask[%d] = 0x%08x ",x,m);
      for(int b=0; b<16; b++) printmaskbit(m,b);
      printf("\n");
    }
    printf("\n");
    printf("trigger    = 0x%x \n",readtrigmask());
    printf("busy       = 0x%x \n",readbusy());
    printf("prescale 1 = %d \n",readpsc(0));
    printf("prescale 2 = %d \n",readpsc(1));
    printf("freq       = %d \n",readfreq());
   
    printf("\n");
    printf("Input counters:\n");
    for(int x=0; x<16; x++) {
      printf("input[%2d] = %d \n",x,counter(x));
    }

    printf("\n");
    printf("Trigger counters:\n");
    for(int x=0; x<4; x++) {
      printf("trigger[%d] all = %12d busy = %12d\n",x,trigall(x),trigbusy(x));
    }
    printf("total      sum = %12d busy = %12d\n",trigsum(),busysum());

  }

  // save configuration
  int save(const char *filename, char **inputnames=NULL) {
    FILE *fp = fopen(filename,"w");
    if(fp==NULL) {
      printf("tl2::save cannot open file %s for writing!\n",filename);
      return -1;
    }

    for(int i=0; i<4; i++) {
      fprintf(fp, "0x%08x\n", readmask(i));
    }
    fprintf(fp, "0x%x\n", readtrigmask());
    fprintf(fp, "0x%x\n", readbusy()&1);
    fprintf(fp, "0x%x\n", readpsc(0));
    fprintf(fp, "0x%x\n", readpsc(1));
    fprintf(fp, "0x%x\n", readfreq());

    if(inputnames!=NULL) {
      for(int i=0; i<16; i++) {
       	fprintf(fp, "%s\n",inputnames[i]);
      }
    }

    fclose(fp);
    return 0;
  }

  int load(const char *filename, char **inputnames=NULL) {
    unsigned int mask;
    FILE *fp = fopen(filename,"r");
    if(fp==NULL) {
      printf("tl2::load cannot open file %s for reading!\n",filename);
      return -1;
    }

    for(int i=0; i<4; i++) {
      fscanf(fp, "%x", &mask);
      writemask(i,mask);
    }

    fscanf(fp, "%x", &mask);
    writetrigmask(mask);

    fscanf(fp, "%x", &mask);
    setbusy(mask);

    fscanf(fp, "%x", &mask);
    writepsc(0,mask);
    fscanf(fp, "%x", &mask);
    writepsc(1,mask);

    fscanf(fp, "%x", &mask);
    writefreq(mask);    
    
    if(inputnames!=NULL) {
      for(int i=0; i<16; i++) fscanf(fp, "%s", inputnames[i]);
    }

    fclose(fp);
    return 0;
  }

 private:
  LE20C* h; // QBUS-SUMMA host
  int n;    // position in crate
};

# endif
