/*
  isa-qbus.c

  PC-QBUS driver

  v1.0 Mar-2004 by Solo
  v1.1 Dec-2006 by Solo - rename to ISA-QBUS to separate from PCI driver
                          add clear function
*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <asm/io.h>
#include <asm/uaccess.h> 

MODULE_AUTHOR("Oleg.Solovyanov@ihep.ru");
MODULE_DESCRIPTION("Driver for ISA-QBUS interface cards");
EXPORT_NO_SYMBOLS;

// global symbols
#define QBUS_IO_RANGE   8
#define QBUS_ADDR_PORT  4
#define QBUS_DATA_PORT  6
#define QBUS_DEVICE_NUM MINOR(filp->f_dentry->d_inode->i_rdev)

// global variables
static unsigned int qbus_major=0;
static char name[2][4]={"iq0","iq1"};
static int io[2]={0x300,0x310};
static int irq[2]={10,11};
static int debug=0;

// board parameters
MODULE_PARM(io,"1-2i");
MODULE_PARM(irq,"1-2i");
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(io,"ISA-QBUS base I/O port (default 0x300)");
MODULE_PARM_DESC(irq,"ISA-QBUS irq line (default 10)");
MODULE_PARM_DESC(debug,"debug level (default 0)");

// interrupt service routine
void qbus_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  if (debug>0) printk("isa-qbus irq=%d interrupt\n",irq);
}

// device open routine
int qbus_open(struct inode *inode, struct file *filp)
{
  int minor=MINOR(inode->i_rdev);
  if(debug>0) printk("isa-qbus%d open(0x%x,%d)\n",minor,io[minor],irq[minor]);
  return 0;
}

// device close routine
int qbus_release(struct inode *inode, struct file *filp)
{
  int minor=MINOR(inode->i_rdev);
  if(debug>0) printk("isa-qbus%d close\n",minor);
  return 0;
}

// address setup
loff_t qbus_llseek(struct file *filp, loff_t offset, int whence) 
{
  int port=io[QBUS_DEVICE_NUM]; // base port

  if(debug>0) printk("isa-qbus%d lseek(0x%x,%d)\n",QBUS_DEVICE_NUM,((short)offset)&0xffff,whence);

  switch(whence) {
  case 0: { // SEEK_SET - set address
    filp->f_pos=offset;
    outw((short)filp->f_pos, port+QBUS_ADDR_PORT); 
    break; 
  }
  case 1: { // SEEK_CUR - clear error status
    outw(0, port+2); 
    break;
  }    
  case 2: { // SEEK_END - device reset
    outw(0, port);
    break;
  }
  default: break;
  }
  
  return filp->f_pos;
}

// read from device
ssize_t qbus_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
  unsigned int port=io[QBUS_DEVICE_NUM]; // base port
  unsigned short w;

  w = inw(port+QBUS_DATA_PORT);
  copy_to_user(buf, &w, 2);

  if(debug>0) printk("isa-qbus%d read=0x%x\n",QBUS_DEVICE_NUM,w);
  
  return 2;
}

// write to device
ssize_t qbus_write(struct file *filp, const char *buf, size_t count, loff_t *offp)
{
  unsigned int port=io[QBUS_DEVICE_NUM]; // base port
  unsigned short w;

  copy_from_user(&w, buf, 2);
  outw(w, port+QBUS_DATA_PORT);

  if(debug>0) printk("isa-qbus%d write(0x%x)\n",QBUS_DEVICE_NUM,w);

  return 2;
}

// available operations
struct file_operations qbus_fops = {
  owner: THIS_MODULE,
  llseek: qbus_llseek,
  read: qbus_read,
  write: qbus_write,
  //  ioctl: qbus_ioctl,
  open: qbus_open,
  release: qbus_release,
};

int init_module(void)
{
  int i;
  printk("isa-qbus.c: v1.1 Dec-2006 by Solo\n");
  i=register_chrdev(0,"iq",&qbus_fops);
  if(i<0) {
    printk("isa-qbus.c: can't get major number\n");
    return i;
  }
  qbus_major=i;

  // allocate hw resources
  for(i=0; i<2; i++) {
    if(io[i]>0) {
      if(!request_region(io[i],QBUS_IO_RANGE,name[i])<0) {
	printk("%s:cannot allocate ISA I/O port %d\n",name[i],io[i]);
	io[i]=0;
	irq[i]=0;
      }
      if(irq[i]>0) {
	if(request_irq(irq[i],qbus_interrupt,0,name[i],NULL)<0) {
	  printk("%s:cannot allocate interrupt line %d\n",name[i],irq[i]);
	  irq[i]=0;
	}
      }      
    }	
  }

  return 0;
}

void cleanup_module(void)
{
  int i;

  printk("isa-qbus.c: module unload\n");

  // release hw resources
  for(i=0; i<2; i++) {
    if(io[i]>0) {
      release_region(io[i],QBUS_IO_RANGE);
    }
    if(irq[i]>0) {
      free_irq(irq[i],NULL);
    }
  }

  unregister_chrdev(qbus_major,"iq");
}

