/*
  isa-qbus.c

  ISA-QBUS driver

  v1.0 Mar-2004 by Solo
  v1.1 Dec-2006 by Solo - rename to ISA-QBUS to separate from PCI driver
                          add clear function
  v1.2 Nov-2007 by Solo - port to kernel 2.6
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/cdev.h>

#include <asm/io.h>
#include <asm/uaccess.h> 

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Oleg.Solovyanov@ihep.ru");
MODULE_DESCRIPTION("Driver for ISA-QBUS interface cards");
MODULE_VERSION("1.2");

// global symbols
#define ISA_QBUS_MAX_DEV    2
#define ISA_QBUS_IO_RANGE   8
#define ISA_QBUS_ADDR_PORT  4
#define ISA_QBUS_DATA_PORT  6
#define ISA_QBUS_DEVICE_NUM MINOR(filp->f_dentry->d_inode->i_rdev)

// global variables
static unsigned int isa_qbus_major=0;
static char name[ISA_QBUS_MAX_DEV][4]={"iq0","iq1"};
static int io[ISA_QBUS_MAX_DEV]={0x300,0x310};
static int irq[ISA_QBUS_MAX_DEV]={10,11};
static struct cdev* cdev[ISA_QBUS_MAX_DEV];
static int debug=0;
static int argc=0;

// board parameters
module_param_array(io,int,argc,0);
module_param_array(irq,int,argc,0);
module_param(debug,int,0);
MODULE_PARM_DESC(io,"ISA-QBUS base I/O port (default 0x300)");
MODULE_PARM_DESC(irq,"ISA-QBUS irq line (default 10)");
MODULE_PARM_DESC(debug,"debug level (default 0)");

// interrupt service routine
irqreturn_t isa_qbus_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
  if (debug>0) printk("isa-qbus irq=%d interrupt\n",irq);
  return IRQ_HANDLED;
}

// device open routine
int isa_qbus_open(struct inode *inode, struct file *filp)
{
  unsigned int minor=iminor(inode);
  if(debug>0) printk(KERN_DEBUG "isa-qbus%d open(0x%x,%d)\n",minor,io[minor],irq[minor]);
  return 0;
}

// device close routine
int isa_qbus_release(struct inode *inode, struct file *filp)
{
  unsigned int minor=iminor(inode);
  if(debug>0) printk(KERN_DEBUG "isa-qbus%d close\n",minor);
  return 0;
}

// address setup
loff_t isa_qbus_llseek(struct file *filp, loff_t offset, int whence) 
{
  int port=io[ISA_QBUS_DEVICE_NUM]; // base port

  if(debug>0) printk("isa-qbus%d lseek(0x%x,%d)\n",ISA_QBUS_DEVICE_NUM,((short)offset)&0xffff,whence);

  switch(whence) {
  case 0: { // SEEK_SET - set address
    filp->f_pos=offset;
    outw((short)filp->f_pos, port+ISA_QBUS_ADDR_PORT); 
    break; 
  }
  case 1: { // SEEK_CUR - clear error status
    outw(0, port+2); 
    break;
  }    
  case 2: { // SEEK_END - device reset
    outw(0, port);
    break;
  }
  default: break;
  }
  
  return filp->f_pos;
}

// read from device
ssize_t isa_qbus_read(struct file *filp, char *buf, size_t count, loff_t *offp)
{
  unsigned int port=io[ISA_QBUS_DEVICE_NUM]; // base port
  unsigned short w;

  w = inw(port+ISA_QBUS_DATA_PORT);
  copy_to_user(buf, &w, 2);

  if(debug>0) printk("isa-qbus%d read=0x%x\n",ISA_QBUS_DEVICE_NUM,w);
  
  return 2;
}

// write to device
ssize_t isa_qbus_write(struct file *filp, const char *buf, size_t count, loff_t *offp)
{
  unsigned int port=io[ISA_QBUS_DEVICE_NUM]; // base port
  unsigned short w;

  copy_from_user(&w, buf, 2);
  outw(w, port+ISA_QBUS_DATA_PORT);

  if(debug>0) printk("isa-qbus%d write(0x%x)\n",ISA_QBUS_DEVICE_NUM,w);

  return 2;
}

// available operations
struct file_operations isa_qbus_fops = {
  .owner = THIS_MODULE,
  .llseek = isa_qbus_llseek,
  .read = isa_qbus_read,
  .write = isa_qbus_write,
  //  .ioctl = isa_qbus_ioctl,
  .open = isa_qbus_open,
  .release = isa_qbus_release,
};

static int __init isa_qbus_init(void)
{
  int i;
  dev_t dev;
  printk("isa-qbus.c: v1.2 Nov-2007 by Solo\n");

  // get the major number and allocate minor numbers
  i=alloc_chrdev_region(&dev,0,ISA_QBUS_MAX_DEV,"iq");
  if(i<0) {
    printk(KERN_WARNING "isa-qbus.c: can't get major number\n");
    return i;
  }
  isa_qbus_major=MAJOR(dev);

  // allocate hw resources
  for(i=0; i<ISA_QBUS_MAX_DEV; i++) {
    if(io[i]>0) {
      if(!request_region(io[i],ISA_QBUS_IO_RANGE,name[i])<0) {
	printk("%s:cannot allocate ISA I/O port %d\n",name[i],io[i]);
	io[i]=0;
	irq[i]=0;
      }
      if(irq[i]>0) {
	if(request_irq(irq[i],isa_qbus_interrupt,0,name[i],NULL)<0) {
	  printk("%s:cannot allocate interrupt line %d\n",name[i],irq[i]);
	  irq[i]=0;
	}
      }      
    }
    // add device structure
    cdev[i]=cdev_alloc();
    cdev[i]->ops=&isa_qbus_fops;
    cdev[i]->owner=THIS_MODULE;	
    cdev_add(cdev[i],MKDEV(isa_qbus_major,i),1);
  }

  return 0;
}

static void __exit isa_qbus_exit(void)
{
  int i;

  printk(KERN_WARNING "isa-qbus.c: module unload\n");

  // release resources
  for(i=0; i<ISA_QBUS_MAX_DEV; i++) {
    if(io[i]>0) {
      if(debug>1) printk("%s:releasing the region 0x%x\n",name[i],io[i]);
      release_region(io[i],ISA_QBUS_IO_RANGE);
    }
    if(irq[i]>0) {
      if(debug>1) printk("%s:releasing the irq %d\n",name[i],irq[i]);
      free_irq(irq[i],NULL);
    }
    cdev_del(cdev[i]);
  }

  unregister_chrdev_region(MKDEV(isa_qbus_major,0),ISA_QBUS_MAX_DEV);
}

// registration
module_init(isa_qbus_init);
module_exit(isa_qbus_exit);
