// LE90m_widget.cpp
//
// 2007 by Solo
// Apr-2009 by Solo - update by timer, load/save
//
#include "le90m_widget.h"

#include <QFileDialog>
#include <QSignalMapper>

static char *le90m_programmable_parameter_name[]={"1 TSSCOPE",
						  "2 TBSGATE",
						  "3 TLED",
						  "4 TGATEES",
						  "5 TINHIBIT",
						  "6 TESSTART",
						  "7 FLED",
						  "8 FTEST"};

static char *le90m_programmable_parameter_unit[]={"s","s","s","s","s","s","Hz","kHz"};

LE90Mwidget::LE90Mwidget(LE90M *le, QWidget *parent) : QWidget(parent)
{
  // copy device pointer
  module = le;

  // create top level layout and box
  QVBoxLayout *mylayout = new QVBoxLayout;
  QGroupBox *mygroupbox = new QGroupBox("LE90M control");
  QVBoxLayout *layout = new QVBoxLayout;
  mygroupbox->setLayout(layout);
  mylayout->addWidget(mygroupbox);

  // add setup parameters
  QGroupBox *sgbox = new QGroupBox("Setup paramaters");
  QVBoxLayout *svbox = new QVBoxLayout;
  setup_parameter[0] = new LE90Msetup("QBUS device",le->host()->host()->device());
  setup_parameter[1] = new LE90Msetup("SUMMA crate",QString::number(le->host()->number()));
  setup_parameter[2] = new LE90Msetup("SUMMA module",QString::number(le->number()));
  svbox->addWidget(setup_parameter[0]);
  svbox->addWidget(setup_parameter[1]);
  svbox->addWidget(setup_parameter[2]);
  sgbox->setLayout(svbox);
  layout->addWidget(sgbox);

  // add enable/disable font panel programming
  QGroupBox *fgbox = new QGroupBox("Front panel programming");
  QHBoxLayout *fhbox = new QHBoxLayout;
  fbenabled = new QRadioButton("Enabled");
  fbdisabled = new QRadioButton("Disabled");
  fhbox->addWidget(fbenabled);
  fhbox->addWidget(fbdisabled);
  fgbox->setLayout(fhbox);
  layout->addWidget(fgbox);

  // add manual parameters 
  QGroupBox *mgbox = new QGroupBox("Manual parameters");
  QVBoxLayout *mvbox = new QVBoxLayout;
  manual_parameter[0] = new LE90Mparameter("TBS","s",true);
  manual_parameter[1] = new LE90Mparameter("TGATE","s",true);
  mvbox->addWidget(manual_parameter[0]);
  mvbox->addWidget(manual_parameter[1]);
  mgbox->setLayout(mvbox);
  layout->addWidget(mgbox);

  // add programmable parameters
  QGroupBox *pgbox = new QGroupBox("Programmable parameters");
  QVBoxLayout *pvbox = new QVBoxLayout;
  for(unsigned int i=0; i<8; i++) {
    programmable_parameter[i]=new LE90Mparameter(le90m_programmable_parameter_name[i],
						 le90m_programmable_parameter_unit[i],
						 true);
    pvbox->addWidget(programmable_parameter[i]);
  }
  pgbox->setLayout(pvbox);
  layout->addWidget(pgbox);

  // add load/save buttons
  QPushButton *bload = new QPushButton("&Load");
  layout->addWidget(bload);
  QPushButton *bsave = new QPushButton("&Save");
  layout->addWidget(bsave);

  setLayout(mylayout);

  // connect buttons
  QObject::connect(bload, SIGNAL(clicked()), this, SLOT(loadFile(void)));
  QObject::connect(bsave, SIGNAL(clicked()), this, SLOT(saveFile(void)));

  // connect parameters
  QSignalMapper *pparMapper = new QSignalMapper(this);
  for(int i=0; i<8; i++) {
    pparMapper->setMapping(programmable_parameter[i],i);
    connect(programmable_parameter[i], SIGNAL(valueChanged(double)), pparMapper, SLOT(map()));  
  }
  connect(pparMapper, SIGNAL(mapped(int)), this, SLOT(changeP(int)));

  QSignalMapper *mparMapper = new QSignalMapper(this);
  for(int i=0; i<2; i++) {
    mparMapper->setMapping(manual_parameter[i],i);
    connect(manual_parameter[i], SIGNAL(valueChanged(double)), mparMapper, SLOT(map()));  
  }
  connect(mparMapper, SIGNAL(mapped(int)), this, SLOT(changeM(int)));

  // update it
  readLE90M();

}

void LE90Mwidget::readLE90M(void)
{
  module->print();

  module->read(&data);
  manual_parameter[0]->setParameter(data.tbs/10.0);
  manual_parameter[1]->setParameter(data.tgate/10.0);

  programmable_parameter[0]->setParameter(data.tsscope/10.0);
  programmable_parameter[1]->setParameter(data.tbsgate/10.0);
  programmable_parameter[2]->setParameter(data.tled/10.0);
  programmable_parameter[3]->setParameter(data.tgatees/10.0);
  programmable_parameter[4]->setParameter(data.tinhibit/10.0);
  programmable_parameter[5]->setParameter(data.tesstart/10.0);
  programmable_parameter[6]->setParameter(100.0/data.fled);
  programmable_parameter[7]->setParameter(100.0/data.ftest);

  data.disable==0?fbenabled->toggle():fbdisabled->toggle();
}

void LE90Mwidget::changeM(int i) {
  if (i==0) module->write_tbs(int(manual_parameter[0]->getParameter()*10));
  if (i==1) module->write_tgate(int(manual_parameter[1]->getParameter()*10));
}

void LE90Mwidget::changeP(int i) {
  if (i>=0 && i<=7) {
    module->write(i,int(programmable_parameter[i]->getParameter()*10));
  }
}

void LE90Mwidget::loadFile(void)
{
  QString fileName = QFileDialog::getOpenFileName(this,
     "Load LE90M configuration file", "/home/roger", "Config Files (*.cfg)");

  if (fileName!=NULL) {
    printf("Loading LE90M configuration from file %s\n",(const char *)fileName.toLocal8Bit().data());
    module->load((const char *)fileName.toLocal8Bit().data());
  }
}

void LE90Mwidget::saveFile(void)
{
  QString fileName = QFileDialog::getSaveFileName(this, 
     "Save LE90M configuration file", "/home/roger", "Config Files (*.cfg)");

  if (fileName!=NULL) {
    printf("Saving LE90M configuration to file %s\n",fileName.toLocal8Bit().data());
    module->save(fileName.toLocal8Bit().data());
  }
}

